import functools
import json
from typing import Union, Type, cast, Any, TypeVar

_cleanups = []


def _tp_cache(func):
    """Internal wrapper caching __getitem__ of generic types with a fallback to
    original function for non-hashable arguments.
    """
    cached = functools.lru_cache()(func)
    _cleanups.append(cached.cache_clear)

    @functools.wraps(func)
    def inner(*args, **kwds):
        try:
            return cached(*args, **kwds)
        except TypeError:
            pass  # All real errors (not unhashable args) are raised below.
        return func(*args, **kwds)

    return inner


class GenericConfigMixin:
    @classmethod
    def __type_check__(cls, instance: Any) -> bool:
        return isinstance(instance, cls)

    @classmethod
    def __typecast__(cls, value: Any) -> Any:
        raise TypeError(f"Object of type {cls.__name__} can not be typecast")


JsonSerializable = Union[int, float, bool, list, dict, str]
ItemType = TypeVar("ItemType", bound=JsonSerializable)


class JsonMeta:
    def __init__(self, origin: Type[ItemType]):
        self.__origin__ = getattr(origin, "__origin__", origin)

    @_tp_cache
    def __getitem__(self, t: Type[ItemType]) -> Type[ItemType]:
        return cast(Type[ItemType], JsonMeta(origin=t))

    def __type_check__(self, instance: Any) -> bool:
        return isinstance(instance, self.__origin__)

    def __typecast__(self, value: Any) -> Any:
        return json.loads(value)


Json = JsonMeta(origin=dict)



import time
import unittest
from unittest.mock import Mock, patch

import requests

import configalchemy.contrib.apollo
from configalchemy.contrib.apollo import ApolloBaseConfig, ConfigException

return_value = {"namespaceName": "tmp", "configurations": {"TEST": "changed"}}


def mock_get(url: str, params: dict = None, timeout: int = 5):
    if "notifications" in url:
        return Mock(
            status_code=200,
            ok=False,
            json=Mock(
                return_value=[{"namespaceName": "application", "notificationId": 1}]
            ),
        )
    else:
        return Mock(status_code=200, ok=True, json=Mock(return_value=return_value))


class ApolloConfigTestCase(unittest.TestCase):
    @patch.object(ApolloBaseConfig, "start_long_poll")
    def test_ApolloConfig(self, start_long_poll):
        requests.get = mock_get

        class DefaultConfig(ApolloBaseConfig):
            TEST = "base"
            ENABLE_LONG_POLL = True
            #: apollo
            APOLLO_SERVER_URL = ""
            APOLLO_APP_ID = ""
            APOLLO_CLUSTER = "default"
            APOLLO_NAMESPACE = "application"

        config = DefaultConfig()
        self.assertEqual("changed", config["TEST"])
        self.assertEqual(1, start_long_poll.call_count)

    @patch.object(requests, "get")
    @patch.object(configalchemy.contrib.apollo, "time_counter")
    @patch.object(time, "sleep")
    def test_long_poll(self, time_sleep, time_counter, requests_get):
        class DefaultConfig(ApolloBaseConfig):
            TEST = "base"
            CONFIGALCHEMY_ENABLE_FUNCTION = False
            #: apollo
            APOLLO_SERVER_URL = ""
            APOLLO_APP_ID = ""
            APOLLO_CLUSTER = "default"
            APOLLO_NAMESPACE = "application"

        count = 0

        def time_counter_side_effect():
            nonlocal count
            if count == 0:
                count += 1
                return 0
            elif count == 1:
                count += 1
                return 99999
            else:
                raise ConfigException("break")

        config = DefaultConfig()
        requests_get.side_effect = mock_get

        self.assertEqual("changed", config.get_from_namespace("TEST", namespace="tmp"))
        return_value["namespaceName"] = "application"
        time_counter.side_effect = time_counter_side_effect
        time_sleep.side_effect = ConfigException("break")

        with self.assertRaises(ConfigException):
            config.long_poll()

        self.assertEqual("changed", config["TEST"])
        self.assertIn("application", config.APOLLO_NOTIFICATION_MAP)

    def test_start_long_poll(self):
        class DefaultConfig(ApolloBaseConfig):
            CONFIGALCHEMY_ENABLE_FUNCTION = False

        config = DefaultConfig()
        with patch("threading.Thread") as MockThread:
            config.start_long_poll()
            MockThread.assert_called_with(target=config.long_poll)


if __name__ == "__main__":
    unittest.main()



import json
import logging
import threading
import time

import requests

from ..configalchemy import BaseConfig, ConfigType

time_counter = time.time


class ConfigException(Exception):
    ...


class ApolloBaseConfig(BaseConfig):
    CONFIGALCHEMY_ENABLE_FUNCTION = True

    #: apollo
    APOLLO_USING_CACHE = False
    APOLLO_SERVER_URL = ""
    APOLLO_APP_ID = ""
    APOLLO_CLUSTER = "default"
    APOLLO_NAMESPACE = "application"
    APOLLO_LONG_POLL_TIMEOUT = 80
    ENABLE_LONG_POLL = False
    APOLLO_NOTIFICATION_MAP: ConfigType = {}

    def __init__(self):
        super().__init__()
        if self.get("ENABLE_LONG_POLL", False):
            self.start_long_poll()

    def get_from_namespace(
        self, key: str, namespace: str = "application", default=None
    ):
        if namespace not in self.APOLLO_NOTIFICATION_MAP:
            self.APOLLO_NAMESPACE = namespace
            self.sync_function()
        return self.APOLLO_NOTIFICATION_MAP[namespace]["data"].get(key, default)

    def start_long_poll(self):
        logging.info("start long poll")
        thread = threading.Thread(target=self.long_poll)
        thread.daemon = True
        thread.start()
        return thread

    def sync_function(self) -> ConfigType:
        route = "configs"
        if self.APOLLO_USING_CACHE:
            route = "configfiles"
        url = (
            f"{self.APOLLO_SERVER_URL}/{route}/{self.APOLLO_APP_ID}/"
            f"{self.APOLLO_CLUSTER}/{self.APOLLO_NAMESPACE}"
        )
        response = requests.get(url)
        if response.ok:
            data = response.json()
            self.APOLLO_NOTIFICATION_MAP.setdefault(data["namespaceName"], {"id": -1})
            self.APOLLO_NOTIFICATION_MAP[data["namespaceName"]]["data"] = data.get(
                "configurations", {}
            )
            logging.debug(f"Got from apollo: {data}")
            return data.get("configurations", {})
        else:
            raise ConfigException("loading config failed")

    def long_poll_from_apollo(self):
        url = f"{self.APOLLO_SERVER_URL}/notifications/v2/"
        notifications = []
        for key, value in self.APOLLO_NOTIFICATION_MAP.items():
            notifications.append({"namespaceName": key, "notificationId": value["id"]})

        r = requests.get(
            url=url,
            params={
                "appId": self.APOLLO_APP_ID,
                "cluster": self.APOLLO_CLUSTER,
                "notifications": json.dumps(notifications, ensure_ascii=False),
            },
            timeout=self.APOLLO_LONG_POLL_TIMEOUT,
        )

        if r.status_code == 304:
            logging.info("Apollo No change, loop...")
        elif r.status_code == 200:
            data = r.json()
            for entry in data:
                logging.info(
                    "%s has changes: notificationId=%d"
                    % (entry["namespaceName"], entry["notificationId"])
                )
                self.APOLLO_NAMESPACE = entry["namespaceName"]
                self.access_config_from_function(
                    priority=self.CONFIGALCHEMY_FUNCTION_VALUE_PRIORITY
                )
                self.APOLLO_NOTIFICATION_MAP[entry["namespaceName"]]["id"] = entry[
                    "notificationId"
                ]
        else:
            logging.info("Sleep...")

    def long_poll(self):
        start_time = time_counter()

        while True:
            try:
                logging.debug("start long poll")
                self.long_poll_from_apollo()
                now = time_counter()
                if now - start_time > 300:
                    for namespace in self.APOLLO_NOTIFICATION_MAP:
                        self.APOLLO_NAMESPACE = namespace
                        self.access_config_from_function(
                            priority=self.CONFIGALCHEMY_FUNCTION_VALUE_PRIORITY
                        )
                    start_time = time_counter()
            except ConfigException:
                time.sleep(5)
            except Exception:
                logging.exception("long poll error")



from json import JSONEncoder
from typing import Any, List, MutableMapping

from configalchemy.field import Field

ConfigType = MutableMapping[str, Any]


class ConfigMetaItem:
    __slots__ = ("priority", "values")

    def __init__(self, priority: int, value: Any):
        self.priority = priority
        self.values = [value]

    def __repr__(self) -> str:
        return f"ConfigMetaItem(priority={self.priority}, value={self.values})"

    def __str__(self) -> str:
        return repr(self)


class ConfigMeta:
    __slots__ = ("field", "items")

    def __init__(self, default_value: Any, field: Field):
        self.field = field
        self.items: List[ConfigMetaItem] = [ConfigMetaItem(0, default_value)]

    @property
    def value(self) -> Any:
        return self.items[-1].values[-1]

    def set(self, priority: int, value: Any) -> None:
        value = self.field.validate(value)
        length = len(self.items)
        for index in range(length - 1, 0, -1):
            if self.items[index - 1].priority < priority:
                item = ConfigMetaItem(priority, value)
                self.items.insert(index, item)
                break
            elif self.items[index - 1].priority == priority:
                self.items[index - 1].values.append(value)
                break
        else:
            if priority == 0:
                self.items[0].values.append(value)
            else:
                item = ConfigMetaItem(priority, value)
                self.items.insert(1, item)

    def __repr__(self) -> str:
        return repr(self.value)

    def __str__(self) -> str:
        return str(self.value)

    def __enter__(self):
        """Prepare to implement optimistic raw lock if necessary"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class ConfigMetaJSONEncoder(JSONEncoder):
    def default(self, o) -> Any:
        if isinstance(o, ConfigMeta):
            return o.value
        super().default(o)



"""Main module."""
import asyncio
import errno
import json
import logging
import os
from typing import (
    Any,
    KeysView,
    List,
    Tuple,
    MutableMapping,
    Dict,
    Optional,
    TypeVar,
    Type,
)
from weakref import ref, ReferenceType

from configalchemy.field import Field
from configalchemy.meta import ConfigMeta, ConfigMetaJSONEncoder

ConfigType = MutableMapping[str, Any]


class BaseConfig(ConfigType):
    """Initialize the :any:`Config` with the Priority::

        ** config from env > config from local file > config from function > default config **

    Example of class-based configuration::

        class DefaultConfig(BaseConfig):
            TEST = "test"

        config = DefaultConfig()
    """

    # The prefix to construct the full environment variable key to access overrode config.
    CONFIGALCHEMY_ENV_PREFIX = ""
    CONFIGALCHEMY_ENVIRONMENT_VALUE_PRIORITY = 20

    # The the filename of the JSON file. This can either be
    # an absolute filename or a filename relative to the
    # `CONFIGALCHEMY_ROOT_PATH`.
    CONFIGALCHEMY_ROOT_PATH = ""
    CONFIGALCHEMY_CONFIG_FILE = ""
    CONFIGALCHEMY_CONFIG_FILE_VALUE_PRIORITY = 10
    # set to ``True`` if you want silent failure for missing files.
    CONFIGALCHEMY_LOAD_FILE_SILENT = False

    # set to ``True`` if you want to override config from function return value.
    CONFIGALCHEMY_ENABLE_FUNCTION = False
    CONFIGALCHEMY_FUNCTION_VALUE_PRIORITY = 0

    # The priority of config['TEST'] = value,
    # config.TEST = value and
    # config.update(TEST=value)
    CONFIGALCHEMY_SETITEM_PRIORITY = 99

    def __init__(self):
        self.meta: Dict[str, ConfigMeta] = {}

        self._setup()

        #: env
        if self.CONFIGALCHEMY_ENV_PREFIX:
            self._from_env()

        #: config file
        if self.CONFIGALCHEMY_CONFIG_FILE:
            self._from_file()

        #: function
        if self.CONFIGALCHEMY_ENABLE_FUNCTION:
            #: Sync
            self.access_config_from_function(
                priority=self.CONFIGALCHEMY_FUNCTION_VALUE_PRIORITY
            )

            # Async
            loop = asyncio.get_event_loop()
            loop.run_until_complete(
                self.access_config_from_coroutine(
                    priority=self.CONFIGALCHEMY_FUNCTION_VALUE_PRIORITY
                )
            )
        global _current_config_ref
        _current_config_ref = ref(self)

    def _setup(self):
        """Setup the default values and field of value from self.
        """
        for key in dir(self):
            if key.isupper():
                self[key] = getattr(self, key)
        return True

    def _from_file(self) -> bool:
        """Updates the values in the config from a JSON file. This function
        behaves as if the JSON object was a dictionary and passed to the
        :meth:`from_mapping` function.
        """
        filename = os.path.join(
            self.CONFIGALCHEMY_ROOT_PATH, self.CONFIGALCHEMY_CONFIG_FILE
        )
        try:
            with open(filename) as f:
                obj = json.load(f)
        except IOError as e:
            if self.CONFIGALCHEMY_LOAD_FILE_SILENT and e.errno in (
                errno.ENOENT,
                errno.EISDIR,
            ):
                return False
            e.strerror = "Unable to load configuration file (%s)" % e.strerror
            raise
        else:
            logging.info(f"Loaded configuration file: {filename}")
            return self.from_mapping(
                obj, priority=self.CONFIGALCHEMY_CONFIG_FILE_VALUE_PRIORITY
            )

    def from_mapping(self, *mappings: ConfigType, priority: int) -> bool:
        """Updates the config like :meth:`update` ignoring items with non-upper
        keys.
        """
        for mapping in mappings:
            for key, value in mapping.items():
                if key.isupper():
                    self._set_value(key, value, priority=priority)
        return True

    def _from_env(self) -> bool:
        """Updates the values in the config from the environment variable.
        """
        for key, value in self.items():
            env_value = os.getenv(f"{self.CONFIGALCHEMY_ENV_PREFIX}{key}")
            if env_value is not None:
                self._set_value(
                    key,
                    env_value,
                    priority=self.CONFIGALCHEMY_ENVIRONMENT_VALUE_PRIORITY,
                )
        return True

    def sync_function(self) -> ConfigType:
        return {}

    async def async_function(self) -> ConfigType:
        return {}

    def access_config_from_function(self, priority: int) -> bool:
        """Updates the values in the config from the sync_access_config_list.
        """
        self.from_mapping(self.sync_function(), priority=priority)
        return True

    async def access_config_from_coroutine(self, priority: int) -> bool:
        """Async updates the values in the config from the async_access_config_list.
        """
        self.from_mapping(await self.async_function(), priority=priority)
        return True

    def _set_value(self, key: str, value: Any, priority: int):
        if key not in self.meta:
            """Setup"""
            self.meta[key] = ConfigMeta(
                default_value=value,
                field=Field(
                    name=key,
                    default_value=value,
                    annotation=getattr(self, "__annotations__", {}).get(key),
                ),
            )
            setattr(self.__class__, key, _ConfigAttribute(key, value))
        else:
            self.meta[key].set(priority=priority, value=value)

    def __getitem__(self, key: str) -> Any:
        """ x.__getitem__(y) <==> x[y] """
        return self.meta[key].value

    def items(self) -> List[Tuple[str, Any]]:  # type: ignore
        return [(key, config_meta.value) for key, config_meta in self.meta.items()]

    def keys(self) -> KeysView[str]:
        return self.meta.keys()

    def __contains__(self, key: object) -> bool:
        return key in self.meta

    def __iter__(self):
        return iter(self.meta)

    def __len__(self) -> int:
        return len(self.meta)

    def __setitem__(self, k, v) -> None:
        self._set_value(k, v, priority=self.CONFIGALCHEMY_SETITEM_PRIORITY)

    def __delitem__(self, key) -> None:
        if len(self.meta[key].items[-1].values) == 1:
            if len(self.meta[key].items) != 1:
                del self.meta[key].items[-1]
        else:
            del self.meta[key].items[-1].values[-1]

    def update(self, __m, **kwargs):
        self.from_mapping(__m, kwargs, priority=self.CONFIGALCHEMY_SETITEM_PRIORITY)

    def get(self, key: str, default=None):
        if key in self.meta:
            return self.meta[key].value
        else:
            return default

    def __bool__(self) -> bool:
        return bool(self.meta)

    def __repr__(self) -> str:
        return repr(self.meta)

    def __str__(self) -> str:
        return repr(self)

    def json(
        self,
        skipkeys: bool = False,
        ensure_ascii: bool = True,
        check_circular: bool = True,
        allow_nan: bool = True,
        sort_keys: bool = False,
        indent: Optional[int] = None,
        separators: Optional[Tuple[str, str]] = None,
    ) -> str:
        return json.dumps(
            self.meta,
            cls=ConfigMetaJSONEncoder,
            skipkeys=skipkeys,
            ensure_ascii=ensure_ascii,
            check_circular=check_circular,
            allow_nan=allow_nan,
            indent=indent,
            separators=separators,
            sort_keys=sort_keys,
        )


class _ConfigAttribute:
    def __init__(self, name: str, default_value: Any):
        self._name = name
        self._default_value = default_value

    def __get__(self, obj: BaseConfig, type=None) -> Any:
        if obj is None:
            return self
        if self._name not in obj:
            return self._default_value
        else:
            return obj[self._name]

    def __set__(self, instance: BaseConfig, value: Any) -> None:
        instance[self._name] = value


_current_config_ref = ref(object)
_CurrentConfigType = TypeVar("_CurrentConfigType", bound=BaseConfig)


def get_current_config(config_type: Type[_CurrentConfigType]) -> _CurrentConfigType:
    # """This API can and should only be used in lazy loading current config instance in the runtime"""
    current_config = _current_config_ref()
    if not isinstance(current_config, config_type):
        raise RuntimeError(f"There is no instance of type {config_type}")
    return current_config



import unittest

from configalchemy import get_current_config, BaseConfig


class GlobalAPITestCase(unittest.TestCase):
    def test_get_current_config(self):
        class DefaultConfig(BaseConfig):
            TEST = "test"

        with self.assertRaises(RuntimeError):
            get_current_config(DefaultConfig)

        with self.assertRaises(RuntimeError):
            DefaultConfig()
            get_current_config(DefaultConfig)
        config = DefaultConfig()
        config.TEST = "inited"
        current_config = get_current_config(DefaultConfig)
        self.assertEqual("inited", current_config.TEST)


if __name__ == "__main__":
    unittest.main()



"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = []

setup_requirements = []

test_requirements = ["requests"]


setup(
    author="GuangTian Li",
    author_email="guangtian_li@qq.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="The Settings and Configuration on ideal practices for app development.",
    install_requires=requirements,
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="configalchemy",
    name="configalchemy",
    packages=find_packages(include=["configalchemy*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/GuangTianLi/configalchemy",
    python_requires=">=3.6.0",
    version="0.2.4",
    zip_safe=False,
    extras_require={"apollo": ["requests"]},
)
